@@ -29,7 +29,8 @@ All operations are performed in a lazy fashion to avoid overloading
 system resources. Conversion to a concrete NumPy ndarray requires an
 explicit method call.
 
-Example:
+For example::
+
     # Wrap two large data sources (e.g. 52000 x 800 x 600).
     measured = OrthoArrayAdapter(netcdf_var_a)
     predicted = OrthoArrayAdapter(netcdf_var_b)
@@ -61,7 +62,11 @@ import numpy as np
 import numpy.ma as ma
 
 
-__version__ = '0.5.0'
+__version__ = '0.6.0'
+
+
+class AxisSupportError(StandardError):
+    """Raised when the operation is not supported over a given axis/axes."""
 
 
 class Engine(object):
@@ -511,6 +516,118 @@ class Array(object):
 
         """
 
+    def _normalise_keys(self, keys):
+        if not isinstance(keys, tuple):
+            keys = (keys,)
+        # This weird check is safe against keys[-1] being an ndarray.
+        if keys and isinstance(keys[-1], type(Ellipsis)):
+            keys = keys[:-1]
+        if len(keys) > self.ndim:
+            raise IndexError('too many keys')
+        for key in keys:
+            if not(isinstance(key, (int, np.integer,
+                                    slice, tuple, np.ndarray))):
+                raise TypeError('invalid index: {!r}'.format(key))
+        return keys
+
+
+class ConstantArray(Array):
+    """
+    An Array which is completely filled with a single value.
+
+    Parameters
+    ----------
+    shape : int or sequence of ints
+        The shape for the new Array.
+    value : obj, optional
+        The value to fill the Array with. Defaults to 0.0.
+    dtype : obj, optional
+        Object to be converted to data type. Default is None which
+        instructs the data type to be determined as the minimum required
+        to hold the given value.
+
+    Returns
+    -------
+    Array
+        An Array entirely filled with ones.
+
+    """
+    def __init__(self, shape, value=0.0, dtype=None):
+        if isinstance(shape, basestring):
+            shape = (shape,)
+        else:
+            try:
+                shape = tuple(shape)
+            except TypeError:
+                shape = (shape,)
+        self._shape = tuple(map(int, shape))
+        data = np.array([value], dtype=dtype)
+        self.value = data[0]
+        self._dtype = data.dtype
+
+    @property
+    def dtype(self):
+        return self._dtype
+
+    @property
+    def shape(self):
+        return self._shape
+
+    def __getitem__(self, keys):
+        keys = self._normalise_keys(keys)
+        shape = _sliced_shape(self.shape, keys)
+        return ConstantArray(shape, self.value, self._dtype)
+
+    def ndarray(self):
+        result = np.empty(self.shape, self._dtype)
+        result.fill(self.value)
+        return result
+
+    def masked_array(self):
+        result = np.ma.empty(self.shape, self._dtype)
+        result.fill(self.value)
+        return result
+
+
+def zeros(shape, dtype=float):
+    """
+    Return an Array which is completely filled with zeros.
+
+    Parameters
+    ----------
+    shape : int or sequence of ints
+        The shape for the new Array.
+    dtype : obj, optional
+        Object to be converted to data type. Default is `float`.
+
+    Returns
+    -------
+    Array
+        An Array entirely filled with zeros.
+
+    """
+    return ConstantArray(shape, dtype=dtype)
+
+
+def ones(shape, dtype=float):
+    """
+    Return an Array which is completely filled with ones.
+
+    Parameters
+    ----------
+    shape : int or sequence of ints
+        The shape for the new Array.
+    dtype : obj, optional
+        Object to be converted to data type. Default is `float`.
+
+    Returns
+    -------
+    Array
+        An Array entirely filled with ones.
+
+    """
+    return ConstantArray(shape, 1, dtype=dtype)
+
 
 class _ArrayAdapter(Array):
     """
@@ -556,7 +673,7 @@ class _ArrayAdapter(Array):
         Raises IndexError/TypeError for invalid keys.
 
         """
-        if isinstance(key, int):
+        if isinstance(key, (int, np.integer)):
             if key >= size or key < -size:
                 msg = 'index {0} is out of bounds for axis {1} with' \
                       ' size {2}'.format(key, axis, size)
@@ -586,7 +703,7 @@ class _ArrayAdapter(Array):
 
         """
         size = len(indices)
-        if isinstance(new_key, int):
+        if isinstance(new_key, (int, np.integer)):
             if new_key >= size or new_key < -size:
                 msg = 'index {0} is out of bounds for axis {1}' \
                       ' with size {2}'.format(new_key, axis, size)
@@ -610,10 +727,7 @@ class _ArrayAdapter(Array):
         return result_key
 
     def __getitem__(self, keys):
-        if not isinstance(keys, tuple):
-            keys = (keys,)
-        if len(keys) > self.ndim:
-            raise IndexError('too many keys')
+        keys = self._normalise_keys(keys)
 
         result_keys = []
         shape = list(self.concrete.shape)
@@ -626,7 +740,7 @@ class _ArrayAdapter(Array):
         while src_keys and new_keys:
             src_size = shape.pop(0)
             src_key = src_keys.pop(0)
-            if isinstance(src_key, int):
+            if isinstance(src_key, (int, np.integer)):
                 # An integer src_key means this dimension has
                 # already been sliced away - it's not visible to
                 # the new keys.
@@ -723,7 +837,8 @@ class NumpyArrayAdapter(_ArrayAdapter):
                 for i, key in tuple_keys:
                     cut_keys[i] = slice(None)
                 array = self.concrete[tuple(cut_keys)]
-                is_scalar = [isinstance(key, int) for key in cut_keys]
+                is_scalar = [isinstance(key, (int, np.integer))
+                             for key in cut_keys]
                 dimensions -= np.cumsum(is_scalar)
             else:
                 # Use ellipsis indexing to ensure we have a real ndarray
@@ -765,6 +880,38 @@ class OrthoArrayAdapter(_ArrayAdapter):
         return array
 
 
+def _pairwise(iterable):
+    """
+    itertools recipe
+    "s -> (s0,s1), (s1,s2), (s2, s3), ...
+
+    """
+    a, b = itertools.tee(iterable)
+    next(b, None)
+    return itertools.izip(a, b)
+
+
+def _groups_of(length, total_length):
+    """
+    Return an iterator of tuples for slicing, in 'length' chunks.
+
+    Parameters
+    ----------
+    length : int
+        Length of each chunk.
+    total_length : int
+        Length of the object we are slicing
+
+    Returns
+    -------
+    iterable of tuples
+        Values defining a slice range resulting in length 'length'.
+
+    """
+    indices = tuple(range(0, total_length, length)) + (None, )
+    return _pairwise(indices)
+
+
 class ArrayStack(Array):
     """
     An Array made from a homogeneous array of other Arrays.
@@ -775,13 +922,21 @@ class ArrayStack(Array):
         item_shape = first_array.shape
         dtype = first_array.dtype
         fill_value = first_array.fill_value
-        fill_value_is_nan = np.isnan(fill_value)
+        if np.issubdtype(dtype, np.floating):
+            def fill_value_ok(array):
+                return array.fill_value == fill_value or (
+                    np.isnan(fill_value) and np.isnan(array.fill_value))
+        else:
+            def fill_value_ok(array):
+                return array.fill_value == fill_value
         for array in stack.flat:
-            both_nan = fill_value_is_nan and np.isnan(array.fill_value)
+            if not isinstance(array, Array):
+                raise ValueError('sub-array must be subclass of Array')
             ok = (array.shape == item_shape and array.dtype == dtype and
-                  (fill_value == array.fill_value or both_nan))
+                  fill_value_ok(array))
             if not ok:
                 raise ValueError('invalid sub-array')
+
         self._stack = stack
         self._item_shape = item_shape
         self._dtype = dtype
@@ -800,16 +955,7 @@ class ArrayStack(Array):
         return self._stack.shape + self._item_shape
 
     def __getitem__(self, keys):
-        if not isinstance(keys, tuple):
-            keys = (keys,)
-        # This weird check is safe against keys[-1] being an ndarray.
-        if isinstance(keys[-1], type(Ellipsis)):
-            keys = keys[:-1]
-        if len(keys) > self.ndim:
-            raise IndexError('too many keys')
-        for key in keys:
-            if not(isinstance(key, (int, slice, tuple, np.ndarray))):
-                raise TypeError('invalid index: {!r}'.format(key))
+        keys = self._normalise_keys(keys)
 
         stack_ndim = self._stack.ndim
         stack_keys = keys[:stack_ndim]
@@ -831,7 +977,7 @@ class ArrayStack(Array):
     def __setitem__(self, keys, value):
         assert len(keys) == self._stack.ndim
         for key in keys:
-            assert isinstance(key, int)
+            assert isinstance(key, (int, np.integer))
         assert isinstance(value, Array), type(value)
         self._stack[keys] = value
 
@@ -849,6 +995,76 @@ class ArrayStack(Array):
             data[index] = masked_array
         return data
 
+    @staticmethod
+    def multidim_array_stack(arrays, shape, order='C'):
+        """
+        Create an N-dimensional ArrayStack from the sequence of Arrays of the
+        same shape.
+
+        Example usage: stacking 6 Arrays, each of shape (768, 1024) to a
+        specified shape (2, 3) will result in an ArrayStack of shape
+        (2, 3, 768, 1024).
+
+        Parameters
+        ----------
+        arrays : sequence of Arrays
+            The sequence of Arrays to be stacked, where each Array must be of
+            the same shape.
+        shape : sequence of ints
+            Shape of the stack, (2, 3) in the above example.
+        order : {'C', 'F'}, optional
+            Use C (C) or FORTRAN (F) index ordering.
+
+        Returns
+        -------
+        ArrayStack
+            With shape corresponding to tuple(stack shape) + tuple(Array.shape)
+            where each Array in the stack must be of the same shape.
+
+        """
+        order = order.lower()
+
+        # Ensure a suitable shape has been specified.
+        size_msg = "total size of new array must be unchanged"
+        try:
+            if np.product(shape) != np.product(arrays.shape):
+                raise ValueError(size_msg)
+            if arrays.ndim > 1:
+                raise ValueError("multidimensional stacks not yet supported")
+        except AttributeError:
+            if np.product(shape) != len(arrays):
+                raise ValueError(size_msg)
+        # Hold the subdivided array
+        subdivided_array = arrays
+
+        # Recursively subdivide to create an ArrayStack with specified shape.
+        for length in shape[::-1]:
+            # Hold the array before this iterations subdivide.
+            array_before_subdivide = subdivided_array
+            subdivided_array = []
+
+            if order == 'f':
+                num = len(array_before_subdivide) // length
+                if length == len(array_before_subdivide):
+                    slc = [slice(None)]
+                else:
+                    slc = [slice(i, None, num) for i in range(num)]
+                for sc in slc:
+                    sub = ArrayStack(np.array(array_before_subdivide[sc],
+                                              dtype=object))
+                    subdivided_array.append(sub)
+            elif order == 'c':
+                for start, stop in _groups_of(length,
+                                              len(array_before_subdivide)):
+                    sub = ArrayStack(np.array(
+                        array_before_subdivide[start:stop], dtype=object))
+                    subdivided_array.append(sub)
+            else:
+                raise TypeError('order not understood')
+        else:
+            # Last iteration, length of the array will be equal to 1.
+            return subdivided_array[0]
+
 
 class LinearMosaic(Array):
     def __init__(self, tiles, axis):
@@ -856,6 +1072,8 @@ class LinearMosaic(Array):
         if tiles.ndim != 1:
             raise ValueError('the tiles array must be 1-dimensional')
         first = tiles[0]
+        if not isinstance(first, Array):
+            raise ValueError('sub-array must be subclass of Array')
         if not(0 <= axis < first.ndim):
             msg = 'invalid axis for {0}-dimensional tiles'.format(first.ndim)
             raise ValueError(msg)
@@ -863,17 +1081,24 @@ class LinearMosaic(Array):
         common_shape = list(first.shape)
         common_dtype = first.dtype
         common_fill_value = first.fill_value
-        common_fill_value_is_nan = np.isnan(common_fill_value)
+        if np.issubdtype(common_dtype, np.floating):
+            def fill_value_ok(array):
+                return array.fill_value == common_fill_value or (
+                    np.isnan(common_fill_value) and np.isnan(array.fill_value))
+        else:
+            def fill_value_ok(array):
+                return array.fill_value == common_fill_value
         del common_shape[axis]
         for tile in tiles[1:]:
+            if not isinstance(tile, Array):
+                raise ValueError('sub-array must be subclass of Array')
             shape = list(tile.shape)
             del shape[axis]
             if shape != common_shape:
                 raise ValueError('inconsistent tile shapes')
             if tile.dtype != common_dtype:
                 raise ValueError('inconsistent tile dtypes')
-            both_nan = common_fill_value_is_nan and np.isnan(tile.fill_value)
-            if (common_fill_value != tile.fill_value and not both_nan):
+            if not fill_value_ok(tile):
                 raise ValueError('inconsistent tile fill_values')
         self._tiles = tiles
         self._axis = axis
@@ -897,10 +1122,7 @@ class LinearMosaic(Array):
         return self._cached_shape
 
     def __getitem__(self, keys):
-        if not isinstance(keys, tuple):
-            keys = (keys,)
-        if len(keys) > self.ndim:
-            raise IndexError('too many keys')
+        keys = self._normalise_keys(keys)
 
         axis = self._axis
         if len(keys) <= axis:
@@ -908,14 +1130,15 @@ class LinearMosaic(Array):
             # then it's safe to just pass the keys to each tile.
             tile = self._tiles[0]
             tiles = [tile[keys] for tile in self._tiles]
-            scalar_keys = filter(lambda key: isinstance(key, int), keys)
+            f = lambda key: isinstance(key, (int, np.integer))
+            scalar_keys = filter(f, keys)
             result = LinearMosaic(tiles, axis - len(scalar_keys))
         else:
             axis_lengths = [tile.shape[axis] for tile in self._tiles]
             offsets = np.cumsum([0] + axis_lengths[:-1])
             splits = offsets - 1
             axis_key = keys[axis]
-            if isinstance(axis_key, int):
+            if isinstance(axis_key, (int, np.integer)):
                 # Find the single relevant tile
                 tile_index = np.searchsorted(splits, axis_key) - 1
                 tile = self._tiles[tile_index]
@@ -938,7 +1161,7 @@ class LinearMosaic(Array):
                 tiles = []
                 tile_slice = list(keys)
                 for tile_index, group_of_pairs in i:
-                    axis_indices = zip(*group_of_pairs)[0]
+                    axis_indices = list(zip(*group_of_pairs))[0]
                     tile = self._tiles[tile_index]
                     axis_indices = np.array(axis_indices)
                     axis_indices -= offsets[tile_index]
@@ -1115,7 +1338,38 @@ class _AggregationStreamsHandler(_StreamsHandler):
         pass
 
 
+class _CountStreamsHandler(_AggregationStreamsHandler):
+    def bootstrap(self, shape):
+        self.current_shape = shape
+        self.running_count = 0
+
+    def finalise(self):
+        count = np.ones(self.current_shape, dtype='i') * self.running_count
+        chunk = Chunk(self.current_keys, count)
+        return chunk
+
+    def process_data(self, data):
+        self.running_count += data.shape[self.axis]
+
+
+class _CountMaskedStreamsHandler(_AggregationStreamsHandler):
+    def bootstrap(self, shape):
+        self.running_count = np.zeros(shape, dtype='i')
+
+    def finalise(self):
+        chunk = Chunk(self.current_keys, self.running_count)
+        return chunk
+
+    def process_data(self, data):
+        self.running_count += np.ma.count(data, axis=self.axis)
+
+
 class _MeanStreamsHandler(_AggregationStreamsHandler):
+    def __init__(self, array, axis, mdtol):
+        # The mdtol argument is not applicable to non-masked arrays
+        # so it is ignored.
+        super(_MeanStreamsHandler, self).__init__(array, axis)
+
     def bootstrap(self, shape):
         self.running_total = np.zeros(shape, dtype=self.array.dtype)
 
@@ -1132,8 +1386,13 @@ class _MeanStreamsHandler(_AggregationStreamsHandler):
 
 
 class _MeanMaskedStreamsHandler(_AggregationStreamsHandler):
+    def __init__(self, array, axis, mdtol):
+        self._mdtol = mdtol
+        super(_MeanMaskedStreamsHandler, self).__init__(array, axis)
+
     def bootstrap(self, shape):
         self.running_count = np.zeros(shape, dtype=self.array.dtype)
+        self.running_masked_count = np.zeros(shape, dtype=self.array.dtype)
         self.running_total = np.zeros(shape, dtype=self.array.dtype)
 
     def finalise(self):
@@ -1144,11 +1403,19 @@ class _MeanMaskedStreamsHandler(_AggregationStreamsHandler):
         # Promote array-scalar to 0-dimensional array.
         if array.ndim == 0:
             array = np.ma.array(array)
+        # Apply masked/missing data threshold (mdtol).
+        if self._mdtol < 1:
+            mask_update = np.true_divide(self.running_masked_count,
+                                         self.running_masked_count +
+                                         self.running_count) > self._mdtol
+            array.mask |= mask_update
+
         chunk = Chunk(self.current_keys, array)
         return chunk
 
     def process_data(self, data):
         self.running_count += np.ma.count(data, axis=self.axis)
+        self.running_masked_count += np.ma.count_masked(data, axis=self.axis)
         self.running_total += np.sum(data, axis=self.axis)
 
 
@@ -1283,7 +1550,7 @@ class _Aggregation(ComputedArray):
     def __init__(self, array, axis,
                  streams_handler_class, masked_streams_handler_class,
                  dtype, kwargs):
-        self._array = array
+        self._array = ensure_array(array)
         self._axis = axis
         self._streams_handler_class = streams_handler_class
         self._masked_streams_handler_class = masked_streams_handler_class
@@ -1305,8 +1572,7 @@ class _Aggregation(ComputedArray):
         return (self._array,)
 
     def __getitem__(self, keys):
-        if not isinstance(keys, tuple):
-            keys = (keys,)
+        keys = self._normalise_keys(keys)
         keys = (slice(None),) + keys
         return _Aggregation(self._array[keys], self._axis,
                             self._streams_handler_class,
@@ -1336,11 +1602,11 @@ def _normalise_axis(axis, array):
     # TypeError/ValueError.
     if axis is None:
         axes = None
-    elif isinstance(axis, int):
+    elif isinstance(axis, (int, np.integer)):
         axes = (axis,)
     elif (isinstance(axis, collections.Iterable) and
             not isinstance(axis, (basestring, collections.Mapping)) and
-            all(map(lambda x: isinstance(x, int), axis))):
+            all(map(lambda x: isinstance(x, (int, np.integer)), axis))):
         axes = tuple(axis)
     else:
         raise TypeError('axis must be None, int, or iterable of ints')
@@ -1351,9 +1617,9 @@ def _normalise_axis(axis, array):
     return axes
 
 
-def mean(a, axis=None):
+def count(a, axis=None):
     """
-    Request the mean of an Array over any number of axes.
+    Count the non-masked elements of the array along the given axis.
 
     .. note:: Currently limited to operating on a single axis.
 
@@ -1370,11 +1636,49 @@ def mean(a, axis=None):
 
     """
     axes = _normalise_axis(axis, a)
-    assert axes is not None and len(axes) == 1
+    if axes is None or len(axes) != 1:
+        msg = "This operation is currently limited to a single axis"
+        raise AxisSupportError(msg)
+    return _Aggregation(a, axes[0],
+                        _CountStreamsHandler, _CountMaskedStreamsHandler,
+                        np.dtype('i'), {})
+
+
+def mean(a, axis=None, mdtol=1):
+    """
+    Request the mean of an Array over any number of axes.
+
+    .. note:: Currently limited to operating on a single axis.
+
+    :param axis: Axis or axes along which the operation is performed.
+                 The default (axis=None) is to perform the operation
+                 over all the dimensions of the input array.
+                 The axis may be negative, in which case it counts from
+                 the last to the first axis.
+                 If axis is a tuple of ints, the operation is performed
+                 over multiple axes.
+    :type axis: None, or int, or iterable of ints.
+    :param float mdtol: Tolerance of missing data. The value in each
+                        element of the resulting array will be masked if the
+                        fraction of masked data contributing to that element
+                        exceeds mdtol. mdtol=0 means no missing data is
+                        tolerated while mdtol=1 will mean the resulting
+                        element will be masked if and only if all the
+                        contributing elements of the source array are masked.
+                        Defaults to 1.
+    :return: The Array representing the requested mean.
+    :rtype: Array
+
+    """
+    axes = _normalise_axis(axis, a)
+    if axes is None or len(axes) != 1:
+        msg = "This operation is currently limited to a single axis"
+        raise AxisSupportError(msg)
     dtype = (np.array([0], dtype=a.dtype) / 1.).dtype
+    kwargs = dict(mdtol=mdtol)
     return _Aggregation(a, axes[0],
                         _MeanStreamsHandler, _MeanMaskedStreamsHandler,
-                        dtype, {})
+                        dtype, kwargs)
 
 
 def std(a, axis=None, ddof=0):
@@ -1399,7 +1703,9 @@ def std(a, axis=None, ddof=0):
 
     """
     axes = _normalise_axis(axis, a)
-    assert axes is not None and len(axes) == 1
+    if axes is None or len(axes) != 1:
+        msg = "This operation is currently limited to a single axis"
+        raise AxisSupportError(msg)
     dtype = (np.array([0], dtype=a.dtype) / 1.).dtype
     return _Aggregation(a, axes[0],
                         _StdStreamsHandler, _StdMaskedStreamsHandler,
@@ -1428,7 +1734,9 @@ def var(a, axis=None, ddof=0):
 
     """
     axes = _normalise_axis(axis, a)
-    assert axes is not None and len(axes) == 1
+    if axes is None or len(axes) != 1:
+        msg = "This operation is currently limited to a single axis"
+        raise AxisSupportError(msg)
     dtype = (np.array([0], dtype=a.dtype) / 1.).dtype
     return _Aggregation(a, axes[0],
                         _VarStreamsHandler, _VarMaskedStreamsHandler,
@@ -1454,6 +1762,9 @@ class _ElementwiseStreamsHandler(_StreamsHandler):
 
 class _Elementwise(ComputedArray):
     def __init__(self, array1, array2, numpy_op, ma_op):
+        array1 = ensure_array(array1)
+        array2 = ensure_array(array2)
+
         # TODO: Broadcasting
         assert array1.shape == array2.shape
         # TODO: Type-promotion
@@ -1476,8 +1787,7 @@ class _Elementwise(ComputedArray):
         return (self._array1, self._array2)
 
     def __getitem__(self, keys):
-        if not isinstance(keys, tuple):
-            keys = (keys,)
+        keys = self._normalise_keys(keys)
         return _Elementwise(self._array1[keys], self._array2[keys],
                             self._numpy_op, self._ma_op)
 
@@ -1534,8 +1844,8 @@ def _sliced_shape(shape, keys):
     """
     sliced_shape = []
     # TODO: Watch out for more keys than shape entries.
-    for size, key in map(None, shape, keys):
-        if isinstance(key, int):
+    for size, key in itertools.izip_longest(shape, keys):
+        if isinstance(key, (int, np.integer)):
             continue
         elif isinstance(key, slice):
             size = len(range(*key.indices(size)))
@@ -1546,3 +1856,21 @@ def _sliced_shape(shape, keys):
             sliced_shape.append(size)
     sliced_shape = tuple(sliced_shape)
     return sliced_shape
+
+
+def ensure_array(array):
+    """
+    Assert that the given array is an Array subclass (or numpy array).
+
+    If the given array is a numpy.ndarray an appropriate NumpyArrayAdapter
+    instance is created, otherwise the passed array must be a subclass of
+    :class:`Array` else a TypeError will be raised.
+
+    """
+    if isinstance(array, np.ndarray):
+        array = NumpyArrayAdapter(array)
+
+    elif not isinstance(array, Array):
+        raise TypeError('The given array should be a `biggus.Array` '
+                        'instance, got {}.'.format(type(array)))
+    return array
